@@ -1,9 +1,10 @@
+import logging
 import ssl
 from dataclasses import dataclass
 from functools import wraps
 from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union
 
-from .. import BrokenResourceError, EndOfStream, aclose_forcefully
+from .. import BrokenResourceError, EndOfStream, aclose_forcefully, get_cancelled_exc_class
 from .._core._typedattr import TypedAttributeSet, typed_attribute
 from ..abc import AnyByteStream, ByteStream, Listener, TaskGroup
 
@@ -145,7 +146,11 @@ class TLSStream(ByteStream):
         await self.transport_stream.aclose()
 
     async def receive(self, max_bytes: int = 65536) -> bytes:
-        return await self._call_sslobject_method(self._ssl_object.read, max_bytes)
+        data = await self._call_sslobject_method(self._ssl_object.read, max_bytes)
+        if not data:
+            raise EndOfStream
+
+        return data
 
     async def send(self, item: bytes) -> None:
         await self._call_sslobject_method(self._ssl_object.write, item)
@@ -182,24 +187,63 @@ class TLSListener(Listener[TLSStream]):
     A convenience listener that wraps another listener and auto-negotiates a TLS session on every
     accepted connection.
 
+    If the TLS handshake times out or raises an exception, :meth:`handle_handshake_error` is
+    called to do whatever post-mortem processing is deemed necessary.
+
     Supports only the :attr:`~TLSAttribute.standard_compatible` extra attribute.
 
     :param Listener listener: the listener to wrap
     :param ssl_context: the SSL context object
     :param standard_compatible: a flag passed through to :meth:`TLSStream.wrap`
+    :param handshake_timeout: time limit for the TLS handshake
+        (passed to :func:`~anyio.fail_after`)
     """
 
     listener: Listener
     ssl_context: ssl.SSLContext
     standard_compatible: bool = True
+    handshake_timeout: float = 30
+
+    @staticmethod
+    async def handle_handshake_error(exc: BaseException, stream: AnyByteStream) -> None:
+        f"""
+        Handle an exception raised during the TLS handshake.
+
+        This method does 3 things:
+
+        #. Forcefully closes the original stream
+        #. Logs the exception (unless it was a cancellation exception) using the ``{__name__}``
+           logger
+        #. Reraises the exception if it was a base exception or a cancellation exception
+
+        :param exc: the exception
+        :param stream: the original stream
+
+        """
+        await aclose_forcefully(stream)
+
+        # Log all except cancellation exceptions
+        if not isinstance(exc, get_cancelled_exc_class()):
+            logging.getLogger(__name__).exception('Error during TLS handshake')
+
+        # Only reraise base exceptions and cancellation exceptions
+        if not isinstance(exc, Exception) or isinstance(exc, get_cancelled_exc_class()):
+            raise
 
     async def serve(self, handler: Callable[[TLSStream], Any],
                     task_group: Optional[TaskGroup] = None) -> None:
         @wraps(handler)
         async def handler_wrapper(stream: AnyByteStream):
-            wrapped_stream = await TLSStream.wrap(stream, ssl_context=self.ssl_context,
-                                                  standard_compatible=self.standard_compatible)
-            await handler(wrapped_stream)
+            from .. import fail_after
+            try:
+                async with fail_after(self.handshake_timeout):
+                    wrapped_stream = await TLSStream.wrap(
+                        stream, ssl_context=self.ssl_context,
+                        standard_compatible=self.standard_compatible)
+            except BaseException as exc:
+                await self.handle_handshake_error(exc, stream)
+            else:
+                await handler(wrapped_stream)
 
         await self.listener.serve(handler_wrapper, task_group)
 
