@@ -9,10 +9,10 @@ from threading import Event, Thread
 import pytest
 
 from anyio import (
-    BusyResourceError, ClosedResourceError, ExceptionGroup, TypedAttributeLookupError, connect_tcp,
-    connect_unix, create_connected_udp_socket, create_event, create_task_group,
-    create_tcp_listener, create_udp_socket, create_unix_listener, getaddrinfo, getnameinfo,
-    move_on_after, wait_all_tasks_blocked)
+    BrokenResourceError, BusyResourceError, ClosedResourceError, ExceptionGroup,
+    TypedAttributeLookupError, connect_tcp, connect_unix, create_connected_udp_socket,
+    create_event, create_task_group, create_tcp_listener, create_udp_socket, create_unix_listener,
+    getaddrinfo, getnameinfo, move_on_after, wait_all_tasks_blocked)
 from anyio.abc.sockets import SocketAttribute
 from anyio.streams.stapled import MultiListener
 
@@ -266,6 +266,27 @@ class TestTCPStream:
         with pytest.raises(ClosedResourceError):
             await stream.send(b'foo')
 
+    async def test_send_after_peer_closed(self, family):
+        def serve_once():
+            client_sock, _ = server_sock.accept()
+            client_sock.close()
+            server_sock.close()
+
+        server_sock = socket.socket(family, socket.SOCK_STREAM)
+        server_sock.settimeout(1)
+        server_sock.bind(('localhost', 0))
+        server_addr = server_sock.getsockname()[:2]
+        server_sock.listen()
+        thread = Thread(target=serve_once, daemon=True)
+        thread.start()
+
+        with pytest.raises(BrokenResourceError):
+            async with await connect_tcp(*server_addr) as stream:
+                for _ in range(1000):
+                    await stream.send(b'foo')
+
+        thread.join()
+
     async def test_connect_tcp_with_tls(self, server_context, client_context, server_sock,
                                         server_addr):
         def serve():
@@ -316,6 +337,7 @@ class TestTCPStream:
 class TestTCPListener:
     async def test_extra_attributes(self, family):
         async with await create_tcp_listener(local_host='localhost', family=family) as multi:
+            assert multi.extra(SocketAttribute.family) == family
             for listener in multi.listeners:
                 raw_socket = listener.extra(SocketAttribute.raw_socket)
                 assert listener.extra(SocketAttribute.family) == family
@@ -389,6 +411,15 @@ class TestTCPListener:
         await multi1.aclose()
         await multi2.aclose()
 
+    async def test_close_from_other_task(self, family):
+        listener = await create_tcp_listener(local_host='localhost', family=family)
+        with pytest.raises(ClosedResourceError):
+            async with create_task_group() as tg:
+                await tg.spawn(listener.serve, lambda stream: None)
+                await wait_all_tasks_blocked()
+                await listener.aclose()
+                await tg.cancel_scope.cancel()
+
 
 @pytest.mark.skipif(sys.platform == 'win32',
                     reason='UNIX sockets are not available on Windows')
@@ -806,6 +837,7 @@ class TestConnectedUDPSocket:
             await udp.send(b'foo')
 
 
+@pytest.mark.network
 async def test_getaddrinfo():
     # IDNA 2003 gets this wrong
     correct = await getaddrinfo('faß.de', 0)
